from __future__ import annotations

import re
from datetime import date, datetime

import pandas as pd
import structlog
from bs4 import BeautifulSoup

from agrobr.exceptions import ParseError

from .models import (
    COLUNAS_SAIDA,
    TICKERS_AGRO,
    UNIDADES,
    parse_numero_br,
    parse_vencimento,
)

logger = structlog.get_logger()

PARSER_VERSION = 1

_RE_ATUALIZADO = re.compile(r"ATUALIZADO EM:\s*(\d{2}/\d{2}/\d{4})")


def _extrair_data_referencia(html: str) -> date | None:
    m = _RE_ATUALIZADO.search(html)
    if not m:
        return None
    return datetime.strptime(m.group(1), "%d/%m/%Y").date()


def _extrair_ticker(texto: str) -> tuple[str, str]:
    parts = texto.split("-", 1)
    ticker = parts[0].strip()
    descricao = parts[1].strip() if len(parts) > 1 else ""
    return ticker, descricao


def parse_ajustes_html(html: str) -> pd.DataFrame:
    data_ref = _extrair_data_referencia(html)
    if data_ref is None:
        logger.info("b3_sem_pregao", reason="ATUALIZADO EM ausente")
        return pd.DataFrame(columns=COLUNAS_SAIDA)

    soup = BeautifulSoup(html, "lxml")
    table = soup.find("table", id="tblDadosAjustes")
    if table is None:
        raise ParseError(
            source="b3",
            parser_version=PARSER_VERSION,
            reason="Tabela tblDadosAjustes nao encontrada",
        )

    rows = table.find_all("tr")  # type: ignore[union-attr]
    if len(rows) < 2:
        raise ParseError(
            source="b3",
            parser_version=PARSER_VERSION,
            reason=f"Tabela com apenas {len(rows)} rows",
        )

    records: list[dict[str, object]] = []
    current_ticker = ""
    current_desc = ""
    in_agro = False

    for row in rows[1:]:
        cells = row.find_all("td")
        if len(cells) < 6:
            continue

        col1 = cells[0].get_text(strip=True)
        if col1:
            current_ticker, current_desc = _extrair_ticker(col1)
            in_agro = current_ticker in TICKERS_AGRO

        if not in_agro:
            continue

        vct_raw = cells[1].get_text(strip=True)
        if not vct_raw:
            continue

        try:
            vct_ano, vct_mes = parse_vencimento(vct_raw)
        except (KeyError, ValueError, IndexError):
            continue

        records.append(
            {
                "data": data_ref,
                "ticker": current_ticker,
                "descricao": current_desc,
                "vencimento_codigo": vct_raw.strip(),
                "vencimento_mes": vct_mes,
                "vencimento_ano": vct_ano,
                "ajuste_anterior": parse_numero_br(cells[2].get_text(strip=True)),
                "ajuste_atual": parse_numero_br(cells[3].get_text(strip=True)),
                "variacao": parse_numero_br(cells[4].get_text(strip=True)),
                "ajuste_por_contrato": parse_numero_br(cells[5].get_text(strip=True)),
                "unidade": UNIDADES.get(current_ticker, ""),
            }
        )

    df = pd.DataFrame(records, columns=COLUNAS_SAIDA)

    if not df.empty:
        df["data"] = pd.to_datetime(df["data"])
        for col in ["ajuste_anterior", "ajuste_atual", "variacao", "ajuste_por_contrato"]:
            df[col] = pd.to_numeric(df[col], errors="coerce")

    logger.info("b3_parse_ok", records=len(df), data_ref=str(data_ref))
    return df



"""API publica do sub-modulo serie historica CONAB."""

from __future__ import annotations

import time
from datetime import UTC, datetime
from typing import Literal, overload

import pandas as pd
import structlog

from agrobr.models import MetaInfo

from . import client
from .parser import PARSER_VERSION, parse_serie_historica, records_to_dataframe

logger = structlog.get_logger()


@overload
async def serie_historica(
    produto: str,
    inicio: int | None = None,
    fim: int | None = None,
    uf: str | None = None,
    *,
    return_meta: Literal[False] = False,
) -> pd.DataFrame: ...


@overload
async def serie_historica(
    produto: str,
    inicio: int | None = None,
    fim: int | None = None,
    uf: str | None = None,
    *,
    return_meta: Literal[True],
) -> tuple[pd.DataFrame, MetaInfo]: ...


async def serie_historica(
    produto: str,
    inicio: int | None = None,
    fim: int | None = None,
    uf: str | None = None,
    return_meta: bool = False,
) -> pd.DataFrame | tuple[pd.DataFrame, MetaInfo]:
    """Obtem serie historica de safras CONAB.

    Busca planilha Excel com dados historicos (area, producao,
    produtividade) por UF desde ~1976 ate a safra mais recente.

    Args:
        produto: Nome do produto (soja, milho, arroz, cafe, etc).
        inicio: Ano inicial (inclusive). Ex: 2020.
        fim: Ano final (inclusive). Ex: 2025.
        uf: Filtrar por UF (ex: 'MT', 'PR'). Se None, retorna todas.
        return_meta: Se True, retorna tupla (DataFrame, MetaInfo).

    Returns:
        DataFrame com colunas: produto, safra, regiao, uf,
        area_plantada_mil_ha, producao_mil_ton, produtividade_kg_ha.

    Example:
        >>> df = await conab.serie_historica("soja", inicio=2020, fim=2025, uf="MT")
        >>> df, meta = await conab.serie_historica("soja", return_meta=True)
    """
    t0 = time.monotonic()

    logger.info(
        "conab_serie_historica_request",
        produto=produto,
        inicio=inicio,
        fim=fim,
        uf=uf,
    )

    xls, metadata = await client.download_xls(produto)

    t1 = time.monotonic()
    records = parse_serie_historica(
        xls=xls,
        produto=produto,
        inicio=inicio,
        fim=fim,
        uf=uf,
    )
    parse_ms = int((time.monotonic() - t1) * 1000)

    df = records_to_dataframe(records)
    fetch_ms = int((time.monotonic() - t0) * 1000)

    logger.info(
        "conab_serie_historica_ok",
        produto=produto,
        records=len(records),
        safras=len(df["safra"].unique()) if not df.empty else 0,
        ufs=len(df["uf"].dropna().unique()) if not df.empty else 0,
    )

    if return_meta:
        meta = MetaInfo(
            source="conab_serie_historica",
            source_url=metadata.get("url", client.SERIES_HISTORICAS_URL),
            source_method="httpx",
            fetched_at=datetime.now(UTC),
            fetch_duration_ms=fetch_ms,
            parse_duration_ms=parse_ms,
            records_count=len(df),
            columns=df.columns.tolist(),
            parser_version=PARSER_VERSION,
            schema_version="1.0",
            attempted_sources=["conab_serie_historica"],
            selected_source="conab_serie_historica",
            fetch_timestamp=datetime.now(UTC),
        )
        return df, meta

    return df


def produtos_disponiveis() -> list[dict[str, str]]:
    """Lista produtos disponiveis para serie historica.

    Returns:
        Lista de dicts com chaves: produto, categoria, url.

    Example:
        >>> produtos = conab.serie_historica.produtos_disponiveis()
        >>> for p in produtos:
        ...     print(p['produto'], p['categoria'])
    """
    return client.list_produtos()



from __future__ import annotations

import pytest

from agrobr.b3.models import (
    B3_CONTRATOS_AGRO,
    MONTH_CODES,
    TICKER_PARA_CONTRATO,
    TICKERS_AGRO,
    UNIDADES,
    parse_numero_br,
    parse_vencimento,
)


class TestContratosAgro:
    def test_all_tickers_unique(self):
        assert len(B3_CONTRATOS_AGRO) == len(set(B3_CONTRATOS_AGRO.values()))

    def test_tickers_agro_matches_values(self):
        assert set(B3_CONTRATOS_AGRO.values()) == TICKERS_AGRO

    def test_ticker_para_contrato_is_inverse(self):
        for nome, ticker in B3_CONTRATOS_AGRO.items():
            assert TICKER_PARA_CONTRATO[ticker] == nome

    def test_all_tickers_have_unidades(self):
        for ticker in TICKERS_AGRO:
            assert ticker in UNIDADES


class TestMonthCodes:
    def test_twelve_months(self):
        assert len(MONTH_CODES) == 12

    def test_values_1_to_12(self):
        assert set(MONTH_CODES.values()) == set(range(1, 13))

    def test_standard_codes(self):
        assert MONTH_CODES["F"] == 1
        assert MONTH_CODES["Z"] == 12
        assert MONTH_CODES["H"] == 3
        assert MONTH_CODES["N"] == 7


class TestParseVencimento:
    def test_g25(self):
        assert parse_vencimento("G25") == (2025, 2)

    def test_h25(self):
        assert parse_vencimento("H25") == (2025, 3)

    def test_z26(self):
        assert parse_vencimento("Z26") == (2026, 12)

    def test_f30(self):
        assert parse_vencimento("F30") == (2030, 1)

    def test_strips_whitespace(self):
        assert parse_vencimento(" K25 ") == (2025, 5)

    def test_invalid_letter_raises(self):
        with pytest.raises(KeyError):
            parse_vencimento("A25")

    def test_invalid_format_raises(self):
        with pytest.raises((ValueError, IndexError)):
            parse_vencimento("")


class TestParseNumeroBr:
    def test_simple_decimal(self):
        assert parse_numero_br("311,45") == 311.45

    def test_thousands_separator(self):
        assert parse_numero_br("1.006,50") == 1006.50

    def test_millions(self):
        assert parse_numero_br("1.066.587,4000") == 1066587.4000

    def test_negative(self):
        assert parse_numero_br("-2,40") == -2.40

    def test_integer_no_separator(self):
        assert parse_numero_br("63") == 63.0

    def test_four_decimals(self):
        assert parse_numero_br("23,7379") == pytest.approx(23.7379)

    def test_empty_returns_none(self):
        assert parse_numero_br("") is None

    def test_dash_returns_none(self):
        assert parse_numero_br("-") is None

    def test_whitespace_stripped(self):
        assert parse_numero_br("  311,45  ") == 311.45



from __future__ import annotations

import httpx
import structlog

from agrobr.constants import HTTPSettings
from agrobr.exceptions import SourceUnavailableError
from agrobr.http.retry import retry_on_status

from .models import COLECAO_ATUAL

logger = structlog.get_logger()

GCS_BASE = "https://storage.googleapis.com/mapbiomas-public/initiatives/brasil"

_settings = HTTPSettings()

TIMEOUT = httpx.Timeout(
    connect=_settings.timeout_connect,
    read=120.0,
    write=_settings.timeout_write,
    pool=_settings.timeout_pool,
)

HEADERS = {"User-Agent": "agrobr (https://github.com/bruno-portfolio/agrobr)"}


def _build_xlsx_url(nivel: str, colecao: int = COLECAO_ATUAL) -> str:
    sufixo = nivel.upper()
    return (
        f"{GCS_BASE}/collection_{colecao}/lulc/statistics/"
        f"MAPBIOMAS_BRAZIL-COL.{colecao}-{sufixo}.xlsx"
    )


async def _fetch_url(url: str) -> bytes:
    async with httpx.AsyncClient(timeout=TIMEOUT, headers=HEADERS, follow_redirects=True) as client:
        logger.debug("mapbiomas_request", url=url)
        response = await retry_on_status(
            lambda: client.get(url),
            source="mapbiomas",
        )

        if response.status_code == 404:
            raise SourceUnavailableError(source="mapbiomas", url=url, last_error="HTTP 404")

        response.raise_for_status()
        return response.content


async def fetch_biome_state(colecao: int = COLECAO_ATUAL) -> tuple[bytes, str]:
    """Baixa XLSX de cobertura por bioma x estado.

    Args:
        colecao: Numero da colecao MapBiomas (default: colecao atual).

    Returns:
        Tupla (bytes_xlsx, url_usada).

    Raises:
        SourceUnavailableError: Se XLSX nao encontrado.
    """
    url = _build_xlsx_url("BIOME_STATE", colecao)
    content = await _fetch_url(url)
    logger.info("mapbiomas_xlsx_found", url=url, size=len(content))
    return content, url


async def fetch_biome_state_municipality(colecao: int = COLECAO_ATUAL) -> tuple[bytes, str]:
    """Baixa XLSX de cobertura por bioma x estado x municipio.

    Args:
        colecao: Numero da colecao MapBiomas (default: colecao atual).

    Returns:
        Tupla (bytes_xlsx, url_usada).

    Raises:
        SourceUnavailableError: Se XLSX nao encontrado.
    """
    url = _build_xlsx_url("BIOME_STATE_MUNICIPALITY", colecao)
    content = await _fetch_url(url)
    logger.info("mapbiomas_xlsx_found", url=url, size=len(content))
    return content, url



from __future__ import annotations

import httpx
import structlog

from agrobr.constants import HTTPSettings
from agrobr.exceptions import SourceUnavailableError
from agrobr.http.retry import retry_on_status

logger = structlog.get_logger()

BASE_URL = "https://dataserver-coids.inpe.br/queimadas/queimadas/focos/csv"

_settings = HTTPSettings()

TIMEOUT = httpx.Timeout(
    connect=_settings.timeout_connect,
    read=60.0,
    write=_settings.timeout_write,
    pool=_settings.timeout_pool,
)

HEADERS = {"User-Agent": "agrobr (https://github.com/bruno-portfolio/agrobr)"}


async def _fetch_url(url: str) -> bytes:
    async with httpx.AsyncClient(timeout=TIMEOUT, headers=HEADERS, follow_redirects=True) as client:
        logger.debug("queimadas_request", url=url)
        response = await retry_on_status(
            lambda: client.get(url),
            source="queimadas",
        )

        if response.status_code == 404:
            raise SourceUnavailableError(source="queimadas", url=url, last_error="HTTP 404")

        response.raise_for_status()
        return response.content


async def fetch_focos_diario(data: str) -> tuple[bytes, str]:
    """Baixa CSV de focos diarios.

    Args:
        data: Data no formato YYYYMMDD.

    Returns:
        Tupla (bytes_csv, url_usada).

    Raises:
        SourceUnavailableError: Se CSV nao encontrado.
    """
    url = f"{BASE_URL}/diario/Brasil/focos_diario_br_{data}.csv"
    content = await _fetch_url(url)
    logger.info("queimadas_csv_found", url=url, size=len(content))
    return content, url


async def fetch_focos_mensal(ano: int, mes: int) -> tuple[bytes, str]:
    """Baixa CSV de focos mensais.

    Args:
        ano: Ano (ex: 2024).
        mes: Mes (1-12).

    Returns:
        Tupla (bytes_csv, url_usada).

    Raises:
        SourceUnavailableError: Se CSV nao encontrado.
    """
    periodo = f"{ano:04d}{mes:02d}"
    url = f"{BASE_URL}/mensal/Brasil/focos_mensal_br_{periodo}.csv"
    content = await _fetch_url(url)
    logger.info("queimadas_csv_found", url=url, size=len(content))
    return content, url



from __future__ import annotations

import httpx
import structlog

from agrobr.constants import HTTPSettings
from agrobr.exceptions import SourceUnavailableError
from agrobr.http.retry import retry_on_status

logger = structlog.get_logger()

BASE_URL = "https://www2.bmf.com.br/pages/portal/bmfbovespa/boletim1/Ajustes1.asp"

_settings = HTTPSettings()

TIMEOUT = httpx.Timeout(
    connect=_settings.timeout_connect,
    read=_settings.timeout_read,
    write=_settings.timeout_write,
    pool=_settings.timeout_pool,
)

HEADERS = {"User-Agent": "agrobr (https://github.com/bruno-portfolio/agrobr)"}


async def fetch_ajustes(data: str) -> tuple[str, str]:
    url = f"{BASE_URL}?txtData={data}"
    async with httpx.AsyncClient(
        timeout=TIMEOUT, headers=HEADERS, follow_redirects=True, verify=False
    ) as http:
        logger.debug("b3_request", url=url)
        response = await retry_on_status(
            lambda: http.get(BASE_URL, params={"txtData": data}),
            source="b3",
        )

        if response.status_code == 404:
            raise SourceUnavailableError(source="b3", url=url, last_error="HTTP 404")

        response.raise_for_status()
        html = response.content.decode("iso-8859-1")
        logger.info("b3_fetch_ok", url=url, size=len(html))
        return html, url



"""Modelos Pydantic para crédito rural BCB/SICOR."""

from __future__ import annotations

import re
from typing import Any

from pydantic import BaseModel, Field, field_validator


class CreditoRural(BaseModel):
    """Registro de crédito rural agregado por município."""

    safra: str = Field(..., description="Safra no formato YYYY/YYYY ou YYYY/YY")
    ano_emissao: int | None = None
    mes_emissao: int | None = Field(None, ge=1, le=12)
    cd_municipio: str | None = Field(None, description="Código IBGE do município (7 dígitos)")
    municipio: str | None = None
    uf: str | None = Field(None, min_length=2, max_length=2)
    produto: str = Field(...)
    finalidade: str = Field(default="custeio")
    valor: float = Field(..., ge=0, description="Valor total do crédito (R$)")
    area_financiada: float | None = Field(None, ge=0, description="Área financiada (ha)")
    qtd_contratos: int | None = Field(None, ge=0, description="Número de contratos")

    @field_validator("uf", mode="before")
    @classmethod
    def normalize_uf(cls, v: Any) -> str | None:
        if v is None:
            return None
        if isinstance(v, str):
            return v.upper().strip()
        return str(v)

    @field_validator("produto", mode="before")
    @classmethod
    def normalize_produto(cls, v: str) -> str:
        return v.lower().strip()

    @field_validator("finalidade", mode="before")
    @classmethod
    def normalize_finalidade(cls, v: str) -> str:
        return v.lower().strip()


# Mapeamento agrobr → SICOR nome do produto
SICOR_PRODUTOS: dict[str, str] = {
    "soja": "SOJA",
    "milho": "MILHO",
    "arroz": "ARROZ",
    "feijao": "FEIJAO",
    "trigo": "TRIGO",
    "algodao": "ALGODAO HERBACEO",
    "cafe": "CAFE",
    "cafe_arabica": "CAFE ARABICA",
    "cafe_conilon": "CAFE CONILON",
    "cana": "CANA-DE-ACUCAR",
    "mandioca": "MANDIOCA",
    "sorgo": "SORGO",
}

SICOR_FINALIDADES: dict[str, str] = {
    "custeio": "CUSTEIO",
    "investimento": "INVESTIMENTO",
    "comercializacao": "COMERCIALIZACAO",
    "industrializacao": "INDUSTRIALIZACAO",
}

# Mapeamento de UF sigla → código IBGE (2 dígitos)
UF_CODES: dict[str, str] = {
    "RO": "11",
    "AC": "12",
    "AM": "13",
    "RR": "14",
    "PA": "15",
    "AP": "16",
    "TO": "17",
    "MA": "21",
    "PI": "22",
    "CE": "23",
    "RN": "24",
    "PB": "25",
    "PE": "26",
    "AL": "27",
    "SE": "28",
    "BA": "29",
    "MG": "31",
    "ES": "32",
    "RJ": "33",
    "SP": "35",
    "PR": "41",
    "SC": "42",
    "RS": "43",
    "MS": "50",
    "MT": "51",
    "GO": "52",
    "DF": "53",
}


def normalize_safra_sicor(safra: str) -> str:
    """Normaliza safra para formato SICOR (YYYY/YYYY).

    Args:
        safra: Safra em formato "2023/24", "2023/2024", ou "2024".

    Returns:
        Safra no formato "2023/2024".
    """
    safra = safra.strip()

    # Formato "2023/2024" — já OK
    if re.match(r"^\d{4}/\d{4}$", safra):
        return safra

    # Formato "2023/24"
    match = re.match(r"^(\d{4})/(\d{2})$", safra)
    if match:
        ano_inicio = int(match.group(1))
        # ano_fim é sempre ano_inicio + 1 para safras agrícolas
        ano_fim = ano_inicio + 1
        return f"{ano_inicio}/{ano_fim}"

    # Formato "2024" — assume "2023/2024"
    if re.match(r"^\d{4}$", safra):
        ano = int(safra)
        return f"{ano - 1}/{ano}"

    return safra


def resolve_produto_sicor(produto: str) -> str:
    """Resolve nome de produto agrobr para nome SICOR.

    Args:
        produto: Nome do produto no formato agrobr.

    Returns:
        Nome do produto no formato SICOR (maiúsculas).
    """
    lower = produto.lower().strip()
    return SICOR_PRODUTOS.get(lower, produto.upper())
